from __future__ import annotations

from collections.abc import Callable, Hashable
from typing import TYPE_CHECKING, Any, cast

import equinox as eqx
import jax
import jaxopt

from dilax.custom_types import Sentinel, _NoValue

__all__ = [
    "JaxOptimizer",
    "Chain",
]


def __dir__():
    return __all__


class JaxOptimizer(eqx.Module):
    """
    Wrapper around `jaxopt` optimizers to make them hashable.
    This allows to pass the optimizer as a parameter to a `jax.jit` function, and setup the optimizer therein.

    Example:

    .. code-block:: python

        optimizer = JaxOptimizer.make(name="GradientDescent", settings={"maxiter": 5})
        # or, e.g.: optimizer = JaxOptimizer.make(name="LBFGS", settings={"maxiter": 10})

        optimizer.fit(fun=nll, init_values=init_values)
    """

    name: str
    _settings: tuple[tuple[str, Hashable], ...]

    def __init__(self, name: str, _settings: tuple[tuple[str, Hashable], ...]) -> None:
        self.name = name
        self._settings = _settings

    @classmethod
    def make(
        cls: type[JaxOptimizer],
        name: str,
        settings: dict[str, Hashable] | Sentinel = _NoValue,
    ) -> JaxOptimizer:
        if settings is _NoValue:
            settings = {}
        if TYPE_CHECKING:
            settings = cast(dict[str, Hashable], settings)
        return cls(name=name, _settings=tuple(settings.items()))

    @property
    def settings(self) -> dict[str, Hashable]:
        return dict(self._settings)

    def solver_instance(self, fun: Callable) -> jaxopt._src.base.Solver:
        return getattr(jaxopt, self.name)(fun=fun, **self.settings)

    def fit(
        self, fun: Callable, init_values: dict[str, jax.Array]
    ) -> tuple[dict[str, jax.Array], Any]:
        values, state = self.solver_instance(fun=fun).run(init_values)
        return values, state


class Chain(eqx.Module):
    """
    Chain multiple optimizers together.
    They probably should have the `maxiter` setting set to a value,
    in order to have a deterministic runtime behaviour.

    Example:

    .. code-block:: python

        opt1 = JaxOptimizer.make(name="GradientDescent", settings={"maxiter": 5})
        opt2 = JaxOptimizer.make(name="LBFGS", settings={"maxiter": 10})

        chain = Chain(opt1, opt2)
        # first 5 steps are minimized with GradientDescent, then 10 steps with LBFGS
        chain.fit(fun=nll, init_values=init_values)
    """

    optimizers: tuple[JaxOptimizer, ...]

    def __init__(self, *optimizers: JaxOptimizer) -> None:
        self.optimizers = optimizers

    def fit(
        self, fun: Callable, init_values: dict[str, jax.Array]
    ) -> tuple[dict[str, jax.Array], Any]:
        values = init_values
        for optimizer in self.optimizers:
            values, state = optimizer.fit(fun=fun, init_values=values)
        return values, state



from __future__ import annotations

import dilax

project = dilax.__name__
copyright = dilax.__copyright__
author = dilax.__author__
version = release = dilax.__version__

language = "en"

templates_path = ["_templates"]
html_static_path = ["_static"]
master_doc = "index"
source_suffix = ".rst"
pygments_style = "sphinx"
add_module_names = False

exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

html_title = f"{project} v{version}"
html_theme = "sphinx_book_theme"
html_theme_options = {
    "logo_only": True,
    "home_page_in_toc": True,
    "show_navbar_depth": 2,
    "show_toc_level": 2,
    "repository_url": "https://github.com/pfackeldey/dilax",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
}

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "autodocsumm",
    "myst_parser",
    "sphinx_copybutton",
    "sphinx_lfs_content",
]

myst_enable_extensions = [
    "colon_fence",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

nitpick_ignore = [
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
]

always_document_param_types = True

autodoc_member_order = "bysource"


def setup(app):
    app.add_css_file("styles_sphinx_book_theme.css")



from __future__ import annotations

import jax.numpy as jnp

import dilax as dlx


class SPlusBModel(dlx.Model):
    def __call__(self, processes: dict, parameters: dict) -> dlx.Result:
        res = dlx.Result()

        mu_modifier = dlx.modifier(
            name="mu", parameter=parameters["mu"], effect=dlx.effect.unconstrained()
        )
        res.add(
            process="signal",
            expectation=mu_modifier(processes[("signal", "nominal")]),
        )

        bkg1_modifier = dlx.compose(
            dlx.modifier(
                name="lnN1",
                parameter=parameters["norm1"],
                effect=dlx.effect.lnN((0.9, 1.1)),
            ),
            dlx.modifier(
                name="shape1_bkg1",
                parameter=parameters["shape1"],
                effect=dlx.effect.shape(
                    up=processes[("background1", "shape_up")],
                    down=processes[("background1", "shape_down")],
                ),
            ),
        )
        res.add(
            process="background1",
            expectation=bkg1_modifier(processes[("background1", "nominal")]),
        )

        bkg2_modifier = dlx.compose(
            dlx.modifier(
                name="lnN2",
                parameter=parameters["norm2"],
                effect=dlx.effect.lnN((0.95, 1.05)),
            ),
            dlx.modifier(
                name="shape1_bkg2",
                parameter=parameters["shape1"],
                effect=dlx.effect.shape(
                    up=processes[("background2", "shape_up")],
                    down=processes[("background2", "shape_down")],
                ),
            ),
        )
        res.add(
            process="background2",
            expectation=bkg2_modifier(processes[("background2", "nominal")]),
        )
        return res


def create_model():
    processes = {
        ("signal", "nominal"): jnp.array([3]),
        ("background1", "nominal"): jnp.array([10]),
        ("background2", "nominal"): jnp.array([20]),
        ("background1", "shape_up"): jnp.array([12]),
        ("background1", "shape_down"): jnp.array([8]),
        ("background2", "shape_up"): jnp.array([23]),
        ("background2", "shape_down"): jnp.array([19]),
    }
    parameters = {
        "mu": dlx.Parameter(value=jnp.array([1.0]), bounds=(0.0, jnp.inf)),
        "norm1": dlx.Parameter(value=jnp.array([0.0])),
        "norm2": dlx.Parameter(value=jnp.array([0.0])),
        "shape1": dlx.Parameter(value=jnp.array([0.0])),
    }

    # return model
    return SPlusBModel(processes=processes, parameters=parameters)


model = create_model()

init_values = model.parameter_values
observation = jnp.array([37])
asimov = model.evaluate().expectation()


# create optimizer (from `jaxopt`)
optimizer = dlx.optimizer.JaxOptimizer.make(
    name="LBFGS",
    settings={"maxiter": 5, "jit": True, "unroll": True},
)



from collections.abc import Callable
from typing import Any

import jax

ArrayLike = jax.typing.ArrayLike
AddOrMul = Callable[[ArrayLike, ArrayLike], jax.Array]


class Sentinel:
    __slots__ = ("repr",)

    def __init__(self, repr: str) -> None:
        self.repr = repr

    def __repr__(self) -> str:
        return self.repr

    __str__ = __repr__


_NoValue: Any = Sentinel("<NoValue>")



from __future__ import annotations

import jax.numpy as jnp
import pytest

import dilax as dlx
from dilax.pdf import Flat, Gauss, Poisson


def test_parameter():
    p = dlx.Parameter(value=jnp.array(1.0), bounds=(jnp.array(0.0), jnp.array(2.0)))

    assert p.value == 1.0
    assert p.update(jnp.array(2.0)).value == 2.0
    assert p.bounds == (0.0, 2.0)

    assert p.boundary_penalty == 0.0
    assert p.update(jnp.array(3.0)).boundary_penalty == jnp.inf


def test_unconstrained():
    p = dlx.Parameter(value=jnp.array(1.0))
    u = dlx.effect.unconstrained()

    assert u.constraint == Flat()
    assert u.scale_factor(p, jnp.array(1.0)) == pytest.approx(1.0)
    assert u.scale_factor(p.update(jnp.array(2.0)), jnp.array(1.0)) == pytest.approx(
        2.0
    )


def test_gauss():
    p = dlx.Parameter(value=jnp.array(0.0))
    g = dlx.effect.gauss(width=jnp.array(1.0))

    assert g.constraint == Gauss(mean=0.0, width=1.0)
    assert g.scale_factor(p, jnp.array(1.0)) == pytest.approx(1.0)
    assert g.scale_factor(p.update(jnp.array(2.0)), jnp.array(1.0)) == pytest.approx(
        3.0
    )


def test_lnN():
    p = dlx.Parameter(value=jnp.array(0.0))
    ln = dlx.effect.lnN(width=(0.9, 1.1))

    assert ln.constraint == Gauss(mean=0.0, width=1.0)
    assert ln.scale_factor(p, jnp.array(1.0)) == pytest.approx(1.0)
    # assert ln.scale_factor(p.update(jnp.array(2.0)), jnp.array(1.0)) == pytest.approx(1.1)


def test_poisson():
    # p = dlx.Parameter(value=jnp.array(0.0))
    po = dlx.effect.poisson(lamb=jnp.array(10))

    assert po.constraint == Poisson(lamb=jnp.array(10))
    # assert po.scale_factor(p, jnp.array(1.0)) == pytest.approx(1.0) # FIXME
    # assert po.scale_factor(p.update(jnp.array(2.0)), jnp.array(1.0)) == pytest.approx(1.1) # FIXME


def test_shape():
    pass


def test_modifier():
    mu = dlx.Parameter(value=jnp.array(1.1))
    norm = dlx.Parameter(value=jnp.array(0.0))

    # unconstrained effect
    m_unconstrained = dlx.modifier(
        name="mu", parameter=mu, effect=dlx.effect.unconstrained()
    )
    assert m_unconstrained(jnp.array([10])) == pytest.approx(11)

    # gauss effect
    m_gauss = dlx.modifier(
        name="norm", parameter=norm, effect=dlx.effect.gauss(jnp.array(0.1))
    )
    assert m_gauss(jnp.array([10])) == pytest.approx(10)

    # lnN effect
    m_lnN = dlx.modifier(
        name="norm", parameter=norm, effect=dlx.effect.lnN(width=(0.9, 1.1))
    )
    assert m_lnN(jnp.array([10])) == pytest.approx(10)

    # poisson effect # FIXME
    # m_poisson = modifier(name="norm", parameter=norm, effect=poisson(jnp.array(10)))
    # assert m_poisson(jnp.array(10)) == pytest.approx(10)

    # shape effect # FIXME
    # effect = shape(up=jnp.array(12), down=jnp.array(8))
    # m_shape = modifier(name="norm", parameter=norm, effect=effect)
    # assert m_shape(jnp.array(10)) == pytest.approx(10)


def test_compose():
    mu = dlx.Parameter(value=jnp.array(1.1))
    norm = dlx.Parameter(value=jnp.array(0.0))

    # unconstrained effect
    m_unconstrained = dlx.modifier(
        name="mu", parameter=mu, effect=dlx.effect.unconstrained()
    )
    # gauss effect
    m_gauss = dlx.modifier(
        name="norm", parameter=norm, effect=dlx.effect.gauss(jnp.array(0.1))
    )

    # compose
    m = dlx.compose(m_unconstrained, m_gauss)

    assert len(m) == 2
    assert m(jnp.array([10])) == pytest.approx(11)



from __future__ import annotations

import equinox as eqx
import jax
import jax.numpy as jnp

from dilax.pdf import HashablePDF
from dilax.util import as1darray

__all__ = [
    "Parameter",
]


def __dir__():
    return __all__


class Parameter(eqx.Module):
    value: jax.Array = eqx.field(converter=as1darray)
    bounds: tuple[jax.Array, jax.Array] = eqx.field(
        static=True, converter=lambda x: tuple(map(as1darray, x))
    )
    constraints: set[HashablePDF] = eqx.field(static=True)

    def __init__(
        self,
        value: jax.Array,
        bounds: tuple[jax.Array, jax.Array] = (as1darray(-jnp.inf), as1darray(jnp.inf)),
    ) -> None:
        self.value = value
        self.bounds = bounds
        self.constraints: set[HashablePDF] = set()

    def update(self, value: jax.Array) -> Parameter:
        return eqx.tree_at(lambda t: t.value, self, value)

    @property
    def boundary_penalty(self) -> jax.Array:
        return jnp.where(
            (self.value < self.bounds[0]) | (self.value > self.bounds[1]),
            jnp.inf,
            0,
        )



import abc
import operator
from typing import TYPE_CHECKING, ClassVar

import equinox as eqx
import jax
import jax.numpy as jnp

from dilax.custom_types import AddOrMul, ArrayLike
from dilax.parameter import Parameter
from dilax.pdf import Flat, Gauss, HashablePDF, Poisson
from dilax.util import as1darray

if TYPE_CHECKING:
    from typing import ClassVar as AbstractClassVar
else:
    from equinox import AbstractClassVar


__all__ = [
    "Effect",
    "unconstrained",
    "gauss",
    "lnN",
    "poisson",
    "shape",
]


def __dir__():
    return __all__


class Effect(eqx.Module):
    apply_op: AbstractClassVar[AddOrMul]

    @property
    @abc.abstractmethod
    def constraint(self) -> HashablePDF:
        ...

    @abc.abstractmethod
    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        ...


class unconstrained(Effect):
    apply_op: ClassVar[AddOrMul] = operator.mul

    @property
    def constraint(self) -> HashablePDF:
        return Flat()

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        return parameter.value


DEFAULT_EFFECT = unconstrained()


class gauss(Effect):
    width: ArrayLike = eqx.field(static=True, converter=as1darray)

    apply_op: ClassVar[AddOrMul] = operator.mul

    def __init__(self, width: ArrayLike) -> None:
        self.width = width

    @property
    def constraint(self) -> HashablePDF:
        return Gauss(mean=0.0, width=1.0)

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        """
        Implementation with (inverse) CDFs is defined as follows:

            .. code-block:: python

                gx = Gauss(mean=1.0, width=self.width)  # type: ignore[arg-type]
                g1 = Gauss(mean=1.0, width=1.0)

                return gx.inv_cdf(g1.cdf(parameter.value + 1))

        But we can use the fast analytical solution instead:

            .. code-block:: python

                return (parameter.value * self.width) + 1

        """
        return (parameter.value * self.width) + 1


class shape(Effect):
    up: jax.Array = eqx.field(converter=as1darray)
    down: jax.Array = eqx.field(converter=as1darray)

    apply_op: ClassVar[AddOrMul] = operator.add

    def __init__(
        self,
        up: jax.Array,
        down: jax.Array,
    ) -> None:
        self.up = up  # +1 sigma
        self.down = down  # -1 sigma

    def vshift(self, sf: jax.Array, sumw: jax.Array) -> jax.Array:
        factor = sf
        dx_sum = self.up + self.down - 2 * sumw
        dx_diff = self.up - self.down

        # taken from https://github.com/nsmith-/jaxfit/blob/8479cd73e733ba35462287753fab44c0c560037b/src/jaxfit/roofit/combine.py#L173C6-L192
        _asym_poly = jnp.array([3.0, -10.0, 15.0, 0.0]) / 8.0

        abs_value = jnp.abs(factor)
        return 0.5 * (
            dx_diff * factor
            + dx_sum
            * jnp.where(
                abs_value > 1.0,
                abs_value,
                jnp.polyval(_asym_poly, factor * factor),
            )
        )

    @property
    def constraint(self) -> HashablePDF:
        return Gauss(mean=0.0, width=1.0)

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        sf = parameter.value
        return self.vshift(sf=sf, sumw=sumw)
        # shift = self.vshift(sf=sf, sumw=sumw)
        # # handle zeros, see: https://github.com/google/jax/issues/5039
        # x = jnp.where(sumw == 0.0, 1.0, sumw)
        # return jnp.where(sumw == 0.0, shift, (x + shift) / x)


class lnN(Effect):
    width: tuple[ArrayLike, ArrayLike] = eqx.field(static=True)

    apply_op: ClassVar[AddOrMul] = operator.mul

    def __init__(
        self,
        width: tuple[ArrayLike, ArrayLike],
    ) -> None:
        # given as (down, up)
        assert isinstance(width, tuple)
        assert len(width) == 2
        self.width = width

    def interpolate(self, parameter: Parameter) -> jax.Array:
        # https://github.com/cms-analysis/HiggsAnalysis-CombinedLimit/blob/be488af288361ef101859a398ae618131373cad7/src/ProcessNormalization.cc#L112-L129
        x = parameter.value
        lo, hi = self.width
        hi = jnp.log(hi)
        lo = jnp.log(lo)
        lo = -lo
        avg = 0.5 * (hi + lo)
        halfdiff = 0.5 * (hi - lo)
        twox = x + x
        twox2 = twox * twox
        alpha = 0.125 * twox * (twox2 * (3 * twox2 - 10.0) + 15.0)
        return jnp.where(
            jnp.abs(x) >= 0.5, jnp.where(x >= 0, hi, lo), avg + alpha * halfdiff
        )

    @property
    def constraint(self) -> HashablePDF:
        return Gauss(mean=0.0, width=1.0)

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        """
        Implementation with (inverse) CDFs is defined as follows:

            .. code-block:: python

                gx = Gauss(mean=jnp.exp(parameter.value), width=width)  # type: ignore[arg-type]
                g1 = Gauss(mean=1.0, width=1.0)

                return gx.inv_cdf(g1.cdf(parameter.value + 1))

        But we can use the fast analytical solution instead:

            .. code-block:: python

                return jnp.exp(parameter.value * self.interpolate(parameter=parameter))

        """
        return jnp.exp(parameter.value * self.interpolate(parameter=parameter))


class poisson(Effect):
    lamb: jax.Array = eqx.field(static=True, converter=as1darray)

    apply_op: ClassVar[AddOrMul] = operator.mul

    def __init__(self, lamb: jax.Array) -> None:
        self.lamb = lamb

    @property
    def constraint(self) -> HashablePDF:
        return Poisson(lamb=self.lamb)

    def scale_factor(self, parameter: Parameter, sumw: jax.Array) -> jax.Array:
        return parameter.value + 1



from typing import TYPE_CHECKING, cast

import equinox as eqx
import jax
import jax.numpy as jnp

from dilax.custom_types import Sentinel, _NoValue
from dilax.model import Model

__all__ = [
    "NLL",
    "Hessian",
    "CovMatrix",
    "SampleToy",
]


def __dir__():
    return __all__


class BaseModule(eqx.Module):
    """
    Base module to hold the `model` and the `observation`.
    """

    model: Model
    observation: jax.Array = eqx.field(converter=jnp.asarray)

    def __init__(self, model: Model, observation: jax.Array) -> None:
        self.model = model
        self.observation = observation


class NLL(BaseModule):
    """
    Negative log-likelihood (NLL).
    """

    def logpdf(self, *args, **kwargs) -> jax.Array:
        return jax.scipy.stats.poisson.logpmf(*args, **kwargs)

    def __call__(self, values: dict | Sentinel = _NoValue) -> jax.Array:
        if values is _NoValue:
            values = self.model.parameter_values
        model = self.model.update(values=values)
        res = model.evaluate()
        nll = jnp.sum(
            self.logpdf(self.observation, res.expectation())
            - self.logpdf(self.observation, self.observation),
            axis=-1,
        )
        # add constraints
        constraints = jax.tree_util.tree_leaves(model.parameter_constraints())
        nll += sum(constraints)
        nll += model.nll_boundary_penalty()
        return -jnp.sum(nll)


class Hessian(BaseModule):
    """
    Hessian matrix.
    """

    NLL: NLL

    def __init__(self, model: Model, observation: jax.Array) -> None:
        super().__init__(model=model, observation=observation)
        self.NLL = NLL(model=model, observation=observation)

    def __call__(self, values: dict | Sentinel = _NoValue) -> jax.Array:
        if values is _NoValue:
            values = self.model.parameter_values
        if TYPE_CHECKING:
            values = cast(dict, values)
        hessian = jax.hessian(self.NLL, argnums=0)(values)
        hessian, _ = jax.tree_util.tree_flatten(hessian)
        hessian = jnp.array(hessian)
        new_shape = len(values)
        return jnp.reshape(hessian, (new_shape, new_shape))


class CovMatrix(Hessian):
    """
    Covariance matrix.
    """

    def __call__(self, values: dict | Sentinel = _NoValue) -> jax.Array:
        if values is _NoValue:
            values = self.model.parameter_values
        hessian = super().__call__(values=values)
        return jnp.linalg.inv(hessian)


class SampleToy(BaseModule):
    """
    Sample a toy from the model.
    """

    CovMatrix: CovMatrix

    def __init__(self, model: Model, observation: jax.Array) -> None:
        super().__init__(model=model, observation=observation)
        self.CovMatrix = CovMatrix(model=model, observation=observation)

    def __call__(
        self,
        values: dict | Sentinel = _NoValue,
        key: jax.Array | Sentinel = _NoValue,
    ) -> dict[str, jax.Array]:
        if values is _NoValue:
            values = self.model.parameter_values
        if key is _NoValue:
            key = jax.random.PRNGKey(1234)
        if TYPE_CHECKING:
            key = cast(jax.Array, key)
        cov = self.CovMatrix(values=values)
        _values, tree_def = jax.tree_util.tree_flatten(
            self.model.update(values=values).parameter_values
        )
        sampled_values = jax.random.multivariate_normal(
            key=key,
            mean=jnp.concatenate(_values),
            cov=cov,
        )
        new_values = jax.tree_util.tree_unflatten(tree_def, sampled_values)
        model = self.model.update(values=new_values)
        return model.evaluate().expectations



from __future__ import annotations

from jax.config import config
from model import init_values, model, observation, optimizer

from dilax.likelihood import NLL

config.update("jax_enable_x64", True)


# create negative log likelihood
nll = NLL(model=model, observation=observation)

# fit
values, state = optimizer.fit(fun=nll, init_values=init_values)

# update model with fitted values
fitted_model = model.update(values=values)



from __future__ import annotations

import abc
import operator
from functools import reduce
from typing import TYPE_CHECKING

import equinox as eqx
import jax
import jax.numpy as jnp

from dilax.custom_types import AddOrMul
from dilax.effect import (
    DEFAULT_EFFECT,
    gauss,
    poisson,
)
from dilax.parameter import Parameter

if TYPE_CHECKING:
    from dilax.effect import Effect

__all__ = [
    "modifier",
    "compose",
    "staterror",
    "autostaterrors",
]


def __dir__():
    return __all__


class ModifierBase(eqx.Module):
    @abc.abstractmethod
    def __call__(self, sumw: jax.Array) -> jax.Array:
        ...


class modifier(ModifierBase):
    """
    Create a new modifier for a given parameter and penalty.

    Example:

    .. code-block:: python

        import jax.numpy as jnp
        import dilax as dlx

        mu = dlx.Parameter(value=1.1, bounds=(0, 100))
        norm = dlx.Parameter(value=0.0, bounds=(-jnp.inf, jnp.inf))

        # create a new parameter and a penalty
        modify = dlx.modifier(name="mu", parameter=mu, effect=dlx.effect.unconstrained())

        # apply the modifier
        modify(jnp.array([10, 20, 30]))
        # -> Array([11., 22., 33.], dtype=float32, weak_type=True),

        # lnN effect
        modify = dlx.modifier(name="norm", parameter=norm, effect=dlx.effect.lnN((0.8, 1.2)))
        modify(jnp.array([10, 20, 30]))

        # poisson effect
        hist = jnp.array([10, 20, 30])
        modify = dlx.modifier(name="norm", parameter=norm, effect=dlx.effect.poisson(hist))
        modify(jnp.array([10, 20, 30]))

        # shape effect
        up = jnp.array([12, 23, 35])
        down = jnp.array([8, 19, 26])
        modify = dlx.modifier(name="norm", parameter=norm, effect=dlx.effect.shape(up, down))
        modify(jnp.array([10, 20, 30]))
    """

    name: str
    parameter: Parameter
    effect: Effect

    def __init__(
        self, name: str, parameter: Parameter, effect: Effect = DEFAULT_EFFECT
    ) -> None:
        self.name = name
        self.parameter = parameter
        self.effect = effect
        self.parameter.constraints.add(self.effect.constraint)

    def scale_factor(self, sumw: jax.Array) -> jax.Array:
        return self.effect.scale_factor(parameter=self.parameter, sumw=sumw)

    def __call__(self, sumw: jax.Array) -> jax.Array:
        op = self.effect.apply_op
        shift = jnp.atleast_1d(self.scale_factor(sumw=sumw))
        shift = jnp.broadcast_to(shift, sumw.shape)
        return op(shift, sumw)  # type: ignore[call-arg]


class compose(ModifierBase):
    """
    Composition of multiple modifiers, i.e.: `(f ∘ g ∘ h)(hist) = f(hist) * g(hist) * h(hist)`
    It behaves like a single modifier, but it is composed of multiple modifiers; it can be arbitrarly nested.

    Example:

    .. code-block:: python

        import jax.numpy as jnp
        import dilax as dlx

        mu = dlx.Parameter(value=1.1, bounds=(0, 100))
        sigma = dlx.Parameter(value=0.1, bounds=(-100, 100))

        # create a new parameter and a composition of modifiers
        composition = dlx.compose(
            dlx.modifier(name="mu", parameter=mu),
            dlx.modifier(name="sigma1", parameter=sigma, effect=dlx.effect.lnN((0.9, 1.1))),
        )

        # apply the composition
        composition(jnp.array([10, 20, 30]))

        # nest compositions
        composition = dlx.compose(
            composition,
            dlx.modifier(name="sigma2", parameter=sigma, effect=dlx.effect.lnN((0.8, 1.2))),
        )

        # jit
        import equinox as eqx

        eqx.filter_jit(composition)(jnp.array([10, 20, 30]))
    """

    modifiers: list[ModifierBase]

    def __init__(self, *modifiers: modifier) -> None:
        self.modifiers = list(modifiers)
        # unroll nested compositions
        _modifiers = []
        for mod in self.modifiers:
            if isinstance(mod, compose):
                _modifiers.extend(mod.modifiers)
            else:
                assert isinstance(mod, modifier)
                _modifiers.append(mod)
        self.modifiers = _modifiers

    def __check_init__(self):
        # check for duplicate names
        names = [m.name for m in self.modifiers]
        duplicates = {name for name in names if names.count(name) > 1}
        if duplicates:
            msg = f"Modifiers need to have unique names, got: {duplicates}"
            raise ValueError(msg)

    def __len__(self) -> int:
        return len(self.modifiers)

    def __call__(self, sumw: jax.Array) -> jax.Array:
        def _prep_shift(modifier: ModifierBase, sumw: jax.Array) -> jax.Array:
            shift = modifier.scale_factor(sumw=sumw)
            shift = jnp.atleast_1d(shift)
            return jnp.broadcast_to(shift, sumw.shape)

        # collect all multiplicative and additive shifts
        shifts: dict[AddOrMul, list] = {operator.mul: [], operator.add: []}
        for m in range(len(self)):
            modifier = self.modifiers[m]
            if modifier.effect.apply_op is operator.mul:
                shifts[operator.mul].append(_prep_shift(modifier, sumw))
            elif modifier.effect.apply_op is operator.add:
                shifts[operator.add].append(_prep_shift(modifier, sumw))

        # calculate the product with for operator.mul
        _mult_fact = reduce(operator.mul, shifts[operator.mul], 1.0)
        # calculate the sum for operator.add
        _add_shift = reduce(operator.add, shifts[operator.add], 0.0)
        # apply
        return _mult_fact * (sumw + _add_shift)


class staterror(ModifierBase):
    """
    Create a staterror (barlow-beeston) modifier which acts on each bin with a different _underlying_ modifier.

    *Caution*: The instantiation of a `staterror` is not compatible with JAX-transformations (e.g. `jax.jit`)!

    Example:

    .. code-block:: python

        import jax.numpy as jnp
        import dilax as dlx

        hist = jnp.array([10, 20, 30])

        p1 = dlx.Parameter(value=1.0)
        p2 = dlx.Parameter(value=0.0)
        p3 = dlx.Parameter(value=0.0)

        # all bins with bin content below 10 (threshold) are treated as poisson, else gauss
        modify = dlx.staterror(
            parameters={1: p1, 2: p2, 3: p3},
            sumw=hist,
            sumw2=hist,
            threshold=10.0,
        )
        modify(hist)
        # -> Array([13.162277, 20.      , 30.      ], dtype=float32)

        # jit
        import equinox as eqx

        fast_modify = eqx.filter_jit(modify)
    """

    name: str = "staterror"
    parameters: dict[str, Parameter]
    sumw: jax.Array
    sumw2: jax.Array
    sumw2sqrt: jax.Array
    widths: jax.Array
    mask: jax.Array
    threshold: float

    def __init__(
        self,
        parameters: dict[str, Parameter],
        sumw: jax.Array,
        sumw2: jax.Array,
        threshold: float,
    ) -> None:
        self.parameters = parameters
        self.sumw = sumw
        self.sumw2 = sumw2
        self.sumw2sqrt = jnp.sqrt(sumw2)
        self.threshold = threshold

        # calculate width
        self.widths = self.sumw2sqrt / self.sumw

        # store if sumw is below threshold
        self.mask = self.sumw < self.threshold

        for i, name in enumerate(self.parameters):
            param = self.parameters[name]
            effect = poisson(self.sumw[i]) if self.mask[i] else gauss(self.widths[i])
            param.constraints.add(effect.constraint)

    def __check_init__(self):
        if not len(self.parameters) == len(self.sumw2) == len(self.sumw):
            msg = (
                f"Length of parameters ({len(self.parameters)}), "
                f"sumw2 ({len(self.sumw2)}) and sumw ({len(self.sumw)}) "
                "must be the same."
            )
            raise ValueError(msg)
        if not self.threshold > 0.0:
            msg = f"Threshold must be >= 0.0, got: {self.threshold}"
            raise ValueError(msg)

    def scale_factor(self, sumw: jax.Array) -> jax.Array:
        from functools import partial

        assert len(sumw) == len(self.parameters) == len(self.sumw2)

        values = jnp.concatenate([param.value for param in self.parameters.values()])
        idxs = jnp.arange(len(sumw))

        # sumw where mask (poisson) else widths (gauss)
        _widths = jnp.where(self.mask, self.sumw, self.widths)

        def _mod(
            value: jax.Array,
            width: jax.Array,
            idx: jax.Array,
            effect: Effect,
        ) -> jax.Array:
            return effect(width).scale_factor(
                parameter=Parameter(value=value),
                sumw=sumw[idx],
            )[0]

        _poisson_mod = partial(_mod, effect=poisson)
        _gauss_mod = partial(_mod, effect=gauss)

        # apply
        return jnp.where(
            self.mask,
            jax.vmap(_poisson_mod)(values, _widths, idxs),
            jax.vmap(_gauss_mod)(values, _widths, idxs),
        )

    def __call__(self, sumw: jax.Array) -> jax.Array:
        # both gauss and poisson behave multiplicative
        op = operator.mul
        sf = self.scale_factor(sumw=sumw)
        return op(jnp.atleast_1d(sf), sumw)


class autostaterrors(eqx.Module):
    class Mode(eqx.Enumeration):
        barlow_beeston_full = (
            "Barlow-Beeston (full) approach: Poisson per process and bin"
        )
        poisson_gauss = "Poisson (Gauss) per process and bin if sumw < (>) threshold"
        barlow_beeston_lite = "Barlow-Beeston (lite) approach"

    sumw: dict[str, jax.Array]
    sumw2: dict[str, jax.Array]
    masks: dict[str, jax.Array]
    threshold: float
    mode: str
    key_template: str = eqx.field(static=True)

    def __init__(
        self,
        sumw: dict[str, jax.Array],
        sumw2: dict[str, jax.Array],
        threshold: float = 10.0,
        mode: str = Mode.barlow_beeston_lite,
        key_template: str = "__staterror_{process}__",
    ) -> None:
        self.sumw = sumw
        self.sumw2 = sumw2
        self.masks = {p: _sumw < threshold for p, _sumw in sumw.items()}
        self.threshold = threshold
        self.mode = mode
        self.key_template = key_template

    def __check_init__(self):
        if jax.tree_util.tree_structure(self.sumw) != jax.tree_util.tree_structure(
            self.sumw2
        ):  # type: ignore[operator]
            msg = (
                "The structure of `sumw` and `sumw2` needs to be identical, got "
                f"`sumw`: {jax.tree_util.tree_structure(self.sumw)}) and "
                f"`sumw2`: {jax.tree_util.tree_structure(self.sumw2)})"
            )
            raise ValueError(msg)
        if not self.threshold > 0.0:
            msg = f"Threshold must be >= 0.0, got: {self.threshold}"
            raise ValueError(msg)
        if not isinstance(self.mode, self.Mode):
            msg = f"Mode must be of type {self.Mode}, got: {self.mode}"
            raise ValueError(msg)

    def prepare(
        self
    ) -> tuple[dict[str, dict[str, Parameter]], dict[str, dict[str, eqx.Partial]]]:
        """
        Helper to automatically create parameters used by `staterror`
        for the initialisation of a `dlx.Model`.

        *Caution*: This function is not compatible with JAX-transformations (e.g. `jax.jit`)!

        Example:

            .. code-block:: python

                import jax.numpy as jnp
                import dilax as dlx

                sumw = {
                    "signal": jnp.array([5, 20, 30]),
                    "background": jnp.array([5, 20, 30]),
                }

                sumw2 = {
                    "signal": jnp.array([5, 20, 30]),
                    "background": jnp.array([5, 20, 30]),
                }


                auto = dlx.autostaterrors(
                    sumw=sumw,
                    sumw2=sumw2,
                    threshold=10.0,
                    mode=dlx.autostaterrors.Mode.barlow_beeston_full,
                )
                parameters, staterrors = auto.prepare()

                # barlow-beeston-lite
                auto2 = dlx.autostaterrors(
                    sumw=sumw,
                    sumw2=sumw2,
                    threshold=10.0,
                    mode=dlx.autostaterrors.Mode.barlow_beeston_lite,
                )
                parameters2, staterrors2 = auto2.prepare()

                # materialize:
                process = "signal"
                pkey = auto.key_template.format(process=process)
                modify = staterrors[pkey](parameters[pkey])
                modified_process = modify(sumw[process])
        """
        import equinox as eqx

        parameters: dict[str, dict[str, Parameter]] = {}
        staterrors: dict[str, dict[str, eqx.Partial]] = {}

        for process, _sumw in self.sumw.items():
            key = self.key_template.format(process=process)
            process_parameters = parameters[key] = {}
            mask = self.masks[process]
            for i in range(len(_sumw)):
                pkey = f"{process}_{i}"
                if self.mode == self.Mode.barlow_beeston_lite and not mask[i]:
                    # we merge all processes into one parameter
                    # for the barlow-beeston-lite approach where
                    # the bin content is above a certain threshold
                    pkey = f"{i}"
                process_parameters[pkey] = Parameter(value=jnp.array(0.0))
            # prepare staterror
            kwargs = {
                "sumw": _sumw,
                "sumw2": self.sumw2[process],
                "threshold": self.threshold,
            }
            if self.mode == self.Mode.barlow_beeston_full:
                kwargs["threshold"] = jnp.inf  # inf -> always poisson
            elif self.mode == self.Mode.barlow_beeston_lite:
                kwargs["sumw"] = jnp.where(
                    mask,
                    _sumw,
                    sum(jax.tree_util.tree_leaves(self.sumw)),
                )
                kwargs["sumw2"] = jnp.where(
                    mask,
                    self.sumw2[process],
                    sum(jax.tree_util.tree_leaves(self.sumw2)),
                )
            staterrors[key] = eqx.Partial(staterror, **kwargs)
        return parameters, staterrors



from __future__ import annotations

import jax.numpy as jnp
import pytest

from dilax.pdf import Flat, Gauss, Poisson


def test_flat():
    pdf = Flat()

    assert pdf.pdf(jnp.array(1.0)) == jnp.array(1.0)
    assert pdf.pdf(jnp.array(2.0)) == jnp.array(1.0)
    assert pdf.pdf(jnp.array(3.0)) == jnp.array(1.0)

    assert pdf.logpdf(jnp.array(1.0)) == jnp.array(0.0)
    assert pdf.logpdf(jnp.array(2.0)) == jnp.array(0.0)
    assert pdf.logpdf(jnp.array(3.0)) == jnp.array(0.0)


def test_gauss():
    pdf = Gauss(mean=0.0, width=1.0)

    assert pdf.pdf(jnp.array(0.0)) == pytest.approx(1.0 / jnp.sqrt(2 * jnp.pi))
    assert pdf.logpdf(jnp.array(0.0)) == pytest.approx(0.0)
    assert pdf.cdf(jnp.array(0.0)) == pytest.approx(0.5)
    assert pdf.inv_cdf(jnp.array(0.5)) == pytest.approx(0.0)
    assert pdf.inv_cdf(pdf.cdf(jnp.array(0.0))) == pytest.approx(0.0)


def test_poisson():
    pdf = Poisson(lamb=jnp.array(10))

    assert pdf.pdf(jnp.array(0)) == pytest.approx(0.12510978)
    assert pdf.logpdf(jnp.array(-0.5)) == pytest.approx(-1.196003)
    assert pdf.cdf(jnp.array(0)) == pytest.approx(0.5830412)
    # assert pdf.inv_cdf(jnp.array(0.5830412)) == pytest.approx(10)
    # assert pdf.inv_cdf(pdf.cdf(jnp.array(10))) == pytest.approx(10)


def test_hashable():
    assert hash(Flat()) == hash(Flat())
    assert hash(Gauss(mean=0.0, width=1.0)) == hash(Gauss(mean=0.0, width=1.0))
    assert hash(Poisson(lamb=jnp.array(10))) == hash(Poisson(lamb=jnp.array(10)))



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, cast

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from dilax.custom_types import Sentinel, _NoValue
from dilax.parameter import Parameter
from dilax.util import deep_update

__all__ = [
    "Result",
    "Model",
]


def __dir__():
    return __all__


class Result(eqx.Module):
    expectations: dict[str, jax.Array]

    def __init__(self) -> None:
        self.expectations = {}

    def add(self, process: str, expectation: jax.Array) -> Result:
        self.expectations[process] = expectation
        return self

    def expectation(self) -> jax.Array:
        return cast(jax.Array, sum(jtu.tree_leaves(self.expectations)))


def _is_parameter(leaf: Any) -> bool:
    return isinstance(leaf, Parameter)


def _is_none_or_is_parameter(leaf: Any) -> bool:
    return leaf is None or _is_parameter(leaf)


class Model(eqx.Module):
    """
    A model describing nuisance parameters, templates (histograms), and how they interact.
    It is requires to implement the `evaluate` method, which returns an `Result` object.

    Example:

    .. code-block:: python

        import equinox as eqx
        import jax
        import jax.numpy as jnp

        import dilax as dlx


        # Define a simple model with two processes and two parameters
        class MyModel(dlx.Model):
            def __call__(self, processes: dict, parameters: dict) -> dlx.Result:
                res = dlx.Result()

                # signal
                mu_mod = dlx.modifier(name="mu", parameter=parameters["mu"], effect=dlx.effect.unconstrained())
                res.add(process="signal", expectation=mu_mod(processes["signal"]))

                # background
                bkg_mod = dlx.modifier(name="sigma", parameter=parameters["sigma"], effect=dlx.effect.lnN(0.2))
                res.add(process="background", expectation=bkg_mod(processes["background"]))
                return res


        # Setup model
        processes = {"signal": jnp.array([10]), "background": jnp.array([50])}
        parameters = {
            "mu": dlx.Parameter(value=jnp.array([1.0]), bounds=(0.0, jnp.inf)),
            "sigma": dlx.Parameter(value=jnp.array([0.0])),
        }

        model = MyModel(processes=processes, parameters=parameters)

        # evaluate the expectation
        model.evaluate().expectation()
        # -> Array([60.], dtype=float32)

        %timeit model.evaluate().expectation()
        # -> 485 µs ± 1.49 µs per loop (mean ± std. dev. of 7 runs, 1,000 loops each)

        # evaluate the expectation *fast*
        @eqx.filter_jit
        def eval(model) -> jax.Array:
            res = model.evaluate()
            return res.expectation()

        eqx.filter_jit(eval)(model)
        # -> Array([60.], dtype=float32)

        %timeit eqx.filter_jit(eval)(model).block_until_ready()
        # -> 202 µs ± 4.87 µs per loop (mean ± std. dev. of 7 runs, 10,000 loops each)
    """

    processes: dict
    parameters: dict[str, Parameter]
    auxiliary: Any

    def __init__(
        self,
        processes: dict,
        parameters: dict,
        auxiliary: Any | Sentinel = _NoValue,
    ) -> None:
        self.processes = processes
        self.parameters = parameters
        if auxiliary is _NoValue:
            auxiliary = {}
        self.auxiliary = auxiliary

    @property
    def parameter_values(self) -> dict:
        return jtu.tree_map(
            lambda l: l.value,  # noqa: E741
            self.parameters,
            is_leaf=_is_parameter,
        )

    def parameter_constraints(self) -> dict:
        def _constraint(param: Parameter) -> jax.Array:
            if param.constraints:
                if len(param.constraints) > 1:
                    msg = f"More than one constraint per parameter is not allowed. Got: {param.constraint}"
                    raise ValueError(msg)
                return next(iter(param.constraints)).logpdf(param.value)
            return jnp.array([0.0])

        return jtu.tree_map(
            _constraint,
            self.parameters,
            is_leaf=_is_parameter,
        )

    def update(
        self,
        processes: dict | Sentinel = _NoValue,
        values: dict | Sentinel = _NoValue,
    ) -> Model:
        if values is _NoValue:
            values = {}
        if processes is _NoValue:
            processes = {}

        if TYPE_CHECKING:
            values = cast(dict, values)
            processes = cast(dict, processes)

        # patch original processes with new ones
        new_processes = deep_update(self.processes, processes)

        # patch original parameters with new ones
        _updates = deep_update(
            jtu.tree_map(lambda _: None, self.parameters, is_leaf=_is_parameter),
            values,
        )

        def _update_params(update: jax.Array | None, param: Parameter) -> Parameter:
            if update is None:
                return param
            return param.update(value=update)

        new_parameters = jtu.tree_map(
            _update_params,
            _updates,
            self.parameters,
            is_leaf=_is_none_or_is_parameter,
        )

        return eqx.tree_at(
            lambda t: (t.processes, t.parameters), self, (new_processes, new_parameters)
        )

    def nll_boundary_penalty(self) -> jax.Array:
        return cast(
            jax.Array,
            sum(
                jtu.tree_leaves(
                    jtu.tree_map(
                        lambda p: p.boundary_penalty,
                        self.parameters,
                        is_leaf=_is_parameter,
                    )
                )
            ),
        )

    @abc.abstractmethod
    def __call__(self, processes: dict, parameters: dict) -> Result:
        ...

    def evaluate(self) -> Result:
        # evaluate the model with its current state
        return self(self.processes, self.parameters)



from __future__ import annotations

import equinox as eqx
import jax
from jax.config import config
from model import init_values, model, observation, optimizer

from dilax.likelihood import NLL, SampleToy

config.update("jax_enable_x64", True)


# create negative log likelihood
nll = NLL(model=model, observation=observation)

# fit
values, state = optimizer.fit(fun=nll, init_values=init_values)

# create sampling method
sample_toy = SampleToy(model=model, observation=observation)
# vectorise and jit
sample_toys = eqx.filter_vmap(in_axes=(None, 0))(eqx.filter_jit(sample_toy))

sample_toy(values, jax.random.PRNGKey(1234))

# sample 10 toys based on fitted parameters
keys = jax.random.split(jax.random.PRNGKey(1234), num=10)
# postfit toys
toys_postfit = sample_toys(values, keys)
# prefit toys
toys_prefit = sample_toys(init_values, keys)



from __future__ import annotations

import collections
import pprint
from collections.abc import Callable, Hashable, Iterable, Mapping
from typing import TYPE_CHECKING, Any, TypeVar, cast

import jax
import jax.numpy as jnp

from dilax.custom_types import ArrayLike, Sentinel, _NoValue

__all__ = [
    "HistDB",
    "FrozenDB",
    "as1darray",
    "dump_hlo_graph",
    "dump_jaxpr",
    "deep_update",
]


def __dir__():
    return __all__


class FrozenKeysView(collections.abc.KeysView):
    """FrozenKeysView that does not print values when repr'ing."""

    def __init__(self, mapping):
        super().__init__(mapping)
        self._mapping = mapping

    def __repr__(self):
        return f"{type(self).__name__}({list(map(_pretty_key, self._mapping.keys()))})"

    __str__ = __repr__


def _pretty_key(key):
    if not isinstance(key, frozenset):
        key = FrozenDB.keyify(key)
    if len(key) == 1:
        return next(iter(key))
    return tuple([_pretty_key(k) for k in key])


def _indent(amount: int, s: str) -> str:
    """Indents `s` with `amount` spaces."""
    prefix = amount * " "
    return "\n".join(prefix + line for line in s.splitlines())


def _pretty_dict(x):
    if not isinstance(x, Mapping):
        return pprint.pformat(x)
    rep = ""
    for key, val in x.items():
        rep += f"{_pretty_key(key)!r}: {_pretty_dict(val)},\n"
    if rep:
        return "{\n" + _indent(2, rep) + "\n}"
    return "{}"


K = TypeVar("K")
V = TypeVar("V")


def _prepare_freeze(xs: Any) -> Any:
    """Deep copy unfrozen dicts to make the dictionary FrozenDict safe."""
    if isinstance(xs, FrozenDB):
        # we can safely ref share the internal state of a FrozenDict
        # because it is immutable.
        return xs._dict
    if not isinstance(xs, dict):
        # return a leaf as is.
        return xs
    # recursively copy dictionary to avoid ref sharing
    return {FrozenDB.keyify(key): _prepare_freeze(val) for key, val in xs.items()}


def _check_no_duplicate_keys(keys: Iterable[Hashable]) -> None:
    keys = list(keys)
    if any(keys.count(x) > 1 for x in keys):
        msg = f"Duplicate keys: {tuple(keys)}, this is not allowed!"
        raise ValueError(msg)


class FrozenDB(Mapping[K, V]):
    """An immutable database-like custom dict.

    Example:

    .. code-block:: python

        hists = HistDB(
            {
                # QCD
                ("QCD", "nominal"): jnp.array([1, 1, 1, 1, 1]),
                ("QCD", "JES", "Up"): jnp.array([1.5, 1.5, 1.5, 1.5, 1.5]),
                ("QCD", "JES", "Down"): jnp.array([0.5, 0.5, 0.5, 0.5, 0.5]),
                # DY
                ("DY", "nominal"): jnp.array([2, 2, 2, 2, 2]),
                ("DY", "JES", "Up"): jnp.array([2.5, 2.5, 2.5, 2.5, 2.5]),
                ("DY", "JES", "Down"): jnp.array([0.7, 0.7, 0.7, 0.7, 0.7]),
            }
        )

        print(hists)
        # -> HistDB({
        #   ('QCD', 'nominal'): Array([1, 1, 1, 1, 1], dtype=int32),
        #   ('QCD', 'Up', 'JES'): Array([1.5, 1.5, 1.5, 1.5, 1.5], dtype=float32),
        #   ('QCD', 'Down', 'JES'): Array([0.5, 0.5, 0.5, 0.5, 0.5], dtype=float32),
        #   ('DY', 'nominal'): Array([2, 2, 2, 2, 2], dtype=int32),
        #   ('DY', 'Up', 'JES'): Array([2.5, 2.5, 2.5, 2.5, 2.5], dtype=float32),
        #   ('DY', 'Down', 'JES'): Array([0.7, 0.7, 0.7, 0.7, 0.7], dtype=float32),
        # })

        print(hists["QCD"])
        # -> HistDB({
        #     'nominal': Array([1, 1, 1, 1, 1], dtype=int32),
        #     ('Up', 'JES'): Array([1.5, 1.5, 1.5, 1.5, 1.5], dtype=float32),
        #     ('Down', 'JES'): Array([0.5, 0.5, 0.5, 0.5, 0.5], dtype=float32),
        # })

        print(hists["JES"])
        # -> HistDB({
        #     ('QCD', 'Up'): Array([1.5, 1.5, 1.5, 1.5, 1.5], dtype=float32),
        #     ('QCD', 'Down'): Array([0.5, 0.5, 0.5, 0.5, 0.5], dtype=float32),
        #     ('DY', 'Up'): Array([2.5, 2.5, 2.5, 2.5, 2.5], dtype=float32),
        #     ('DY', 'Down'): Array([0.7, 0.7, 0.7, 0.7, 0.7], dtype=float32),
        # })

        # It's jit-compatible:
        def foo(hists):
            return (hists["QCD", "nominal"] + 1.2) ** 2

        print(jax.jit(foo)(hists))
        # -> Array([4.84, 4.84, 4.84, 4.84, 4.84], dtype=float32, weak_type=True)
    """

    __slots__ = ("_dict",)

    if TYPE_CHECKING:
        _dict: dict[frozenset, Any]

    @staticmethod
    def keyify(keyish: Any) -> frozenset:
        if not isinstance(keyish, tuple | list | set | frozenset):
            keyish = (keyish,)
        _check_no_duplicate_keys(keyish)
        keyish = frozenset(keyish)
        assert not any(isinstance(key, set) for key in keyish)
        return keyish

    def __init__(
        self,
        xs: Mapping | Sentinel = _NoValue,
        __unsafe_skip_copy__: bool = False,
    ) -> None:
        # make sure the dict is as
        if xs is _NoValue:
            xs = {}
        data = dict(cast(Mapping, xs))
        if __unsafe_skip_copy__:
            self._dict = data
        else:
            self._dict = _prepare_freeze(data)

    def __getitem__(self, key) -> Any:
        key = self.keyify(key)
        if key in self._dict:
            return self._dict[key]
        ret = self.__class__({k - key: v for k, v in self.items() if key <= k})
        if not ret:
            raise KeyError(key)
        return ret

    def __setitem__(self, key, value) -> None:
        msg = f"{type(self).__name__} is immutable."
        raise ValueError(msg)

    def __contains__(self, key) -> bool:
        key = self.keyify(key)
        return key in self._dict

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    def keys(self) -> FrozenKeysView:
        return FrozenKeysView(self._dict)

    def values(self):
        return self._dict.values()

    def items(self):
        for key in self._dict:
            yield (key, self[key])

    def only(self, *keys) -> FrozenDB:
        return self.__class__({key: self[key] for key in keys})

    def subset(self, *keys) -> FrozenDB:
        new = {}
        for key in keys:
            new.update({k: v for k, v in self.items() if self.keyify(key) <= k})
        return self.__class__(new)

    def copy(self) -> FrozenDB:
        return self.__class__(self)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({_pretty_dict(self._dict)})"

    def as_compact_dict(self):
        return {"/".join(sorted(map(str, k))): v for k, v in self.items()}


def _flatten(tree):
    return (tuple(tree.values()), tuple(tree.keys()))


def _make_unflatten(cls: type[FrozenDB]) -> Callable:
    def _unflatten(keys, values):
        return cls(dict(zip(keys, values, strict=True)), __unsafe_skip_copy__=True)

    return _unflatten


class HistDB(FrozenDB):
    ...


# then we register them with jax as a PyTree
for cls in HistDB, FrozenDB:
    jax.tree_util.register_pytree_node(
        cls,
        _flatten,
        _make_unflatten(cls),
    )


def as1darray(x: ArrayLike) -> jax.Array:
    """
    Converts `x` to a 1d array.

    Example:

    .. code-block:: python

        import jax.numpy as jnp


        as1darray(1.0)
        # -> Array([1.], dtype=float32, weak_type=True)

        as1darray(jnp.array(1.0))
        # -> Array([1.], dtype=float32, weak_type=True)
    """

    return jnp.atleast_1d(jnp.asarray(x))


def dump_jaxpr(fun: Callable, *args: Any, **kwargs: Any) -> str:
    """Helper function to dump the Jaxpr of a function.

    Example:

    .. code-block:: python

        import jax
        import jax.numpy as jnp

        def f(x: jax.Array) -> jax.Array:
            return jnp.sin(x) ** 2 + jnp.cos(x) ** 2

        x = jnp.array([1.0, 2.0, 3.0])

        print(dump_jaxpr(f, x))
        # -> { lambda ; a:f32[3]. let
        #        b:f32[3] = sin a              # []
        #        c:f32[3] = integer_pow[y=2] b # []
        #        d:f32[3] = cos a              # []
        #        e:f32[3] = integer_pow[y=2] d # []
        #        f:f32[3] = add c e            # []
        #      in (f,) }
    """
    jaxpr = jax.make_jaxpr(fun)(*args, **kwargs)
    return jaxpr.pretty_print(name_stack=True)


def dump_hlo_graph(fun: Callable, *args: Any, **kwargs: Any) -> str:
    """
    Helper to dump the HLO graph of a function as a `dot` graph.

    Example:

    .. code-block:: python

        import jax
        import jax.numpy as jnp

        import path


        def f(x: jax.Array) -> jax.Array:
            return x + 1.0

        x = jnp.array([1.0, 2.0, 3.0])

        # dump dot graph to file
        filepath = pathlib.Path('graph.gv')
        filepath.write_text(dump_hlo_graph(f, x), encoding='ascii')
    """
    return jax.xla_computation(fun)(*args, **kwargs).as_hlo_dot_graph()


def deep_update(
    mapping: dict[K, Any],
    new_mapping: dict[K, Any],
) -> dict[K, Any]:
    updated_mapping = mapping.copy()
    for k, v in new_mapping.items():
        if (
            k in updated_mapping
            and isinstance(updated_mapping[k], dict)
            and isinstance(v, dict)
        ):
            updated_mapping[k] = deep_update(updated_mapping[k], v)
        else:
            updated_mapping[k] = v
    return updated_mapping



"""
dilax: Differentiable (binned) likelihoods in JAX.
"""
from __future__ import annotations

__author__ = "Peter Fackeldey"
__email__ = "peter.fackeldey@rwth-aachen.de"
__copyright__ = "Copyright 2023, Peter Fackeldey"
__credits__ = ["Peter Fackeldey"]
__contact__ = "https://github.com/pfackeldey/dilax"
__license__ = "BSD-3-Clause"
__status__ = "Development"
__version__ = "0.1.6"


# expose public API

__all__ = [
    "__version__",
    "effect",
    "ipy_util",
    "likelihood",
    "optimizer",
    "pdf",
    "util",
    # explicitely expose some classes
    "Model",
    "Result",
    "Parameter",
    "modifier",
    "staterror",
    "autostaterrors",
    "compose",
]


def __dir__():
    return __all__


from dilax import (  # noqa: E402
    effect,
    ipy_util,
    likelihood,
    optimizer,
    pdf,
    util,
)
from dilax.model import Model, Result  # noqa: E402
from dilax.modifier import (  # noqa: E402
    autostaterrors,
    compose,
    modifier,
    staterror,
)
from dilax.parameter import Parameter  # noqa: E402



from __future__ import annotations

from functools import partial

import equinox as eqx
import jax
import jax.numpy as jnp
from jax.config import config
from model import asimov, model, optimizer

from dilax import Model
from dilax.likelihood import NLL
from dilax.optimizer import JaxOptimizer

config.update("jax_enable_x64", True)


def nll_profiling(
    value_name: str,
    scan_points: jax.Array,
    model: Model,
    observation: jax.Array,
    optimizer: JaxOptimizer,
    fit: bool,
) -> jax.Array:
    # define single fit for a fixed parameter of interest (poi)
    @partial(jax.jit, static_argnames=("value_name", "optimizer", "fit"))
    def fixed_poi_fit(
        value_name: str,
        scan_point: jax.Array,
        model: Model,
        observation: jax.Array,
        optimizer: JaxOptimizer,
        fit: bool,
    ) -> jax.Array:
        # fix theta into the model
        model = model.update(values={value_name: scan_point})
        init_values = model.parameter_values
        init_values.pop(value_name, 1)
        # minimize
        nll = eqx.filter_jit(NLL(model=model, observation=observation))
        if fit:
            values, _ = optimizer.fit(fun=nll, init_values=init_values)
        else:
            values = model.parameter_values
        return nll(values=values)

    # vectorise for multiple fixed values (scan points)
    fixed_poi_fit_vec = jax.vmap(
        fixed_poi_fit, in_axes=(None, 0, None, None, None, None)
    )
    return fixed_poi_fit_vec(
        value_name, scan_points, model, observation, optimizer, fit
    )


# profile the NLL around starting point of `0`
scan_points = jnp.r_[-1.9:2.0:0.1]

profile_postfit = nll_profiling(
    value_name="norm1",
    scan_points=scan_points,
    model=model,
    observation=asimov,
    optimizer=optimizer,
    fit=True,
)



from typing import Any

import jax.numpy as jnp

from dilax.custom_types import ArrayLike
from dilax.model import Model

__all__ = ["interactive"]


def __dir__():
    return __all__


def interactive(model: Model) -> None:
    import ipywidgets as widgets
    import matplotlib.pyplot as plt

    def slider(v: ArrayLike) -> widgets.FloatSlider:
        return widgets.FloatSlider(min=v - 2, max=v + 2, step=0.01, value=v)

    fig, ax = plt.subplots()

    expectation = model.evaluate().expectation()
    bins = jnp.arange(expectation.size)

    art = ax.bar(bins, expectation, color="gray")

    @widgets.interact(
        **{name: slider(param.value) for name, param in model.parameters.items()}
    )
    def update(**kwargs: Any) -> None:
        m = model.update(values=kwargs)
        res = m.evaluate()

        expectation = res.expectation()
        print("Expectation:", expectation)
        print("Constraint (logpdf):", m.parameter_constraints())

        nonlocal art
        art.remove()

        art = ax.bar(bins, expectation, color="gray")

    ax.set_xticks(bins)
    ax.set_xticklabels(list(map(str, bins)))
    ax.set_xlabel(r"Bin #")
    ax.set_ylabel(r"S+B model")
    plt.tight_layout()
    plt.show()



from __future__ import annotations

import equinox as eqx
from model import model

# see: https://docs.kidger.site/equinox/api/serialisation/

# save processes and parameters
eqx.tree_serialise_leaves("some_filename.eqx", model)

# load processes and parameters back into model
model_loaded = eqx.tree_deserialise_leaves("some_filename.eqx", model)



from __future__ import annotations

import jax

from dilax.util import FrozenDB, as1darray


def get_frozendb():
    return FrozenDB(
        {
            # QCD
            ("a", "b"): 1,
            ("a", "d", "e"): 2,
            ("a", "d", "f"): 3,
            # DY
            ("g", "b"): 4,
            ("g", "d", "e"): 5,
            ("g", "d", "f"): 6,
        }
    )


def test_frozendb_len():
    db = get_frozendb()

    assert len(db) == 6


def test_frozendb_getitem():
    db = get_frozendb()

    assert db["a"]["b"] == 1
    assert db["a", "b"] == 1
    assert db["b"] == FrozenDB({"a": 1, "g": 4})


def test_frozendb_jitcompatible():
    db = get_frozendb()

    @jax.jit
    def fun(db):
        return (db["a", "b"] + 1) ** 2

    assert fun(db) == 4


def test_as1darray():
    arr = as1darray(jax.numpy.array(1.0))

    assert isinstance(arr, jax.Array)
    assert arr.ndim == 1



from __future__ import annotations

import equinox as eqx
from jax.config import config
from model import init_values, model, observation, optimizer

import dilax as dlx

config.update("jax_enable_x64", True)

# create negative log likelihood
nll = dlx.likelihood.NLL(model=model, observation=observation)

# fit
params, state = optimizer.fit(fun=nll, init_values=init_values)

# gradients of nll of fitted model
fast_grad_nll = eqx.filter_jit(eqx.filter_grad(nll))
grads = fast_grad_nll(params)
# gradients of nll of fitted model only wrt to `mu`
# basically: pass the parameters dict of which you want the gradients
params_ = {k: v for k, v in params.items() if k == "mu"}
grad_mu = fast_grad_nll(params_)

# hessian + cov_matrix of nll of fitted model
hessian = eqx.filter_jit(dlx.likelihood.Hessian(model=model, observation=observation))()

# covariance matrix of fitted model
covmatrix = eqx.filter_jit(
    dlx.likelihood.CovMatrix(model=model, observation=observation)
)()



from __future__ import annotations

from functools import partial

import jax
import jaxopt
import pytest

from dilax.optimizer import JaxOptimizer


def test_jaxoptimizer():
    opt = JaxOptimizer.make(name="GradientDescent", settings={"maxiter": 5})

    assert opt.name == "GradientDescent"
    assert opt.settings == {"maxiter": 5}

    assert isinstance(opt.solver_instance(fun=lambda x: x), jaxopt.GradientDescent)

    # jit compatibility
    @partial(jax.jit, static_argnums=0)
    def f(optimizer):
        @jax.jit
        def fun(x):
            return (x - 2.0) ** 2

        init_values = 1.0
        values, _ = optimizer.fit(fun=fun, init_values=init_values)
        return values

    assert f(opt) == pytest.approx(2.0)



from __future__ import annotations

from abc import abstractmethod

import equinox as eqx
import jax
import jax.numpy as jnp

__all__ = [
    "HashablePDF",
    "Flat",
    "Gauss",
    "Poisson",
]


def __dir__():
    return __all__


class HashablePDF(eqx.Module):
    @abstractmethod
    def __hash__(self) -> int:
        ...

    @abstractmethod
    def logpdf(self, x: jax.Array) -> jax.Array:
        ...

    @abstractmethod
    def pdf(self, x: jax.Array) -> jax.Array:
        ...

    @abstractmethod
    def cdf(self, x: jax.Array) -> jax.Array:
        ...

    @abstractmethod
    def inv_cdf(self, x: jax.Array) -> jax.Array:
        ...


class Flat(HashablePDF):
    def __hash__(self):
        return hash(self.__class__)

    def logpdf(self, x: jax.Array) -> jax.Array:
        return jnp.array([0.0])

    def pdf(self, x: jax.Array) -> jax.Array:
        return jnp.array([1.0])

    def cdf(self, x: jax.Array) -> jax.Array:
        return jnp.array([1.0])

    def inv_cdf(self, x: jax.Array) -> jax.Array:
        msg = "Flat distribution has no inverse CDF."
        raise ValueError(msg)


class Gauss(HashablePDF):
    mean: float = eqx.field(static=True)
    width: float = eqx.field(static=True)

    def __init__(self, mean: float, width: float) -> None:
        self.mean = mean
        self.width = width

    def __hash__(self):
        return hash(self.__class__) ^ hash((self.mean, self.width))

    def logpdf(self, x: jax.Array) -> jax.Array:
        logpdf_max = jax.scipy.stats.norm.logpdf(
            self.mean, loc=self.mean, scale=self.width
        )
        unnormalized = jax.scipy.stats.norm.logpdf(x, loc=self.mean, scale=self.width)
        return unnormalized - logpdf_max

    def pdf(self, x: jax.Array) -> jax.Array:
        return jax.scipy.stats.norm.pdf(x, loc=self.mean, scale=self.width)

    def cdf(self, x: jax.Array) -> jax.Array:
        return jax.scipy.stats.norm.cdf(x, loc=self.mean, scale=self.width)

    def inv_cdf(self, x: jax.Array) -> jax.Array:
        return jax.scipy.stats.norm.ppf(x, loc=self.mean, scale=self.width)


class Poisson(HashablePDF):
    lamb: jax.Array = eqx.field(static=True)

    def __init__(self, lamb: jax.Array) -> None:
        self.lamb = lamb

    def __hash__(self):
        return hash(self.__class__) ^ hash(str(self.lamb))  # is this a safe hash??

    def logpdf(self, x: jax.Array) -> jax.Array:
        logpdf_max = jax.scipy.stats.poisson.logpmf(self.lamb, mu=self.lamb)
        unnormalized = jax.scipy.stats.poisson.logpmf((x + 1) * self.lamb, mu=self.lamb)
        return unnormalized - logpdf_max

    def pdf(self, x: jax.Array) -> jax.Array:
        return jax.scipy.stats.poisson.pmf((x + 1) * self.lamb, mu=self.lamb)

    def cdf(self, x: jax.Array) -> jax.Array:
        return jax.scipy.stats.poisson.cdf((x + 1) * self.lamb, mu=self.lamb)

    def inv_cdf(self, x: jax.Array) -> jax.Array:
        # see: https://num.pyro.ai/en/stable/tutorials/truncated_distributions.html?highlight=poisson%20inverse#5.3-Example:-Left-truncated-Poisson
        def cond_fn(val):
            n, cdf = val
            return jnp.any(cdf < x)

        def body_fn(val):
            n, cdf = val
            n_new = jnp.where(cdf < x, n + 1, n)
            return n_new, self.cdf(n_new)

        start = jnp.zeros_like(x)
        cdf_start = self.cdf(start)
        n, _ = jax.lax.while_loop(cond_fn, body_fn, (start, cdf_start))
        return n.astype(jnp.result_type(int))
